
/******************************************************************************
* MODULE     : keyw_retrieve.cc
* DESCRIPTION: The 'retrieve' keyword
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <genalias.hh>
#include <gentable.hh>
namespace mmx {

void
evaluator_rep::kw_retrieve_1 () {
  CHECK_SYNC (MMX_RETRIEVE);
  evaluate_arg (car(trace)[1], KW_RETRIEVE_2);
  // FIXME: should explicitly ask for generator
}

void
evaluator_rep::kw_retrieve_2 () {
  CHECK_SYNC (MMX_RETRIEVE);
  // INPUT: (retrieve_2, generator, cont, ...)
  value   val = as_value (car (rem));
  generic type= val->type;
  if (is_func (type, ID_GENERATOR, 1)) {
    list gen= as_list (as_genalias (val->val) -> get ());
    if (nil (gen)) error (car (trace), "No more values in generator");
    else {
      gentable gen_env  = as_gentable (car (gen));
      list     gen_trace= as_list (cadr (gen));
      list     gen_cont = cddr (gen);
      list     back     = cons (env, trace, KW_RETRIEVE_3, cdr (rem));
      // CONTINUATION for retrieval: (cont, ...)
      gen_env->local_write (CONT_YIELD)= back;
      change_env (gen_env);
      cur= car (gen_cont);
      rem= cdr (gen_cont);
      // OUTPUT: (gen_cont, ...)
      trace= gen_trace;
    }
  }
  else error (car(trace)[1], "Not a generator");
}

void
evaluator_rep::kw_retrieve_3 () {
  // INPUT: (retrieve_3, ret_val, cont, ...)
  if (car (rem) == VAL_DONE)
    error (car (trace), "No more values in generator");
  else {
    cur= cadr (rem);
    rem= cons (car (rem), cddr (rem));
    // OUTPUT: (cont, ret_val, ...)
    trace= cdr (trace);
  }
}

} // namespace mmx
