
/******************************************************************************
* MODULE     : array.cc
* DESCRIPTION: arrays
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <array.hh>
namespace mmx {

/******************************************************************************
* Routines intern to the array class
******************************************************************************/

static inline int
round_length (int n) {
  if (n<6) return n;
  register int i=8;
  while (n>i) i<<=1;
  return i;
}

array_rep::array_rep(int n2):
  n (n2), a ((n==0)? ((generic*) NULL): (new generic [round_length(n)])) {}

array::array (generic* a, int n) {
  register int i;
  rep= new array_rep (n);
  for (i=0; i<n; i++)
    rep->a[i]=a[i];
}

array::array (generic g1) {
  rep= new array_rep (1);
  rep->a[0]= g1;
}

array::array (generic g1, generic g2) {
  rep= new array_rep (2);
  rep->a[0]= g1;
  rep->a[1]= g2;
}

array::array (generic g1, generic g2, generic g3) {
  rep= new array_rep (3);
  rep->a[0]= g1;
  rep->a[1]= g2;
  rep->a[2]= g3;
}

array::array (generic g1, generic g2, generic g3, generic g4) {
  rep= new array_rep (4);
  rep->a[0]= g1;
  rep->a[1]= g2;
  rep->a[2]= g3;
  rep->a[3]= g4;
}

void
array_rep::resize (register int n2) {
  register int l= round_length (n);
  register int m= round_length (n2);
  if (m!=l) {
    if (m!=0) {
      register int i,k=(n2<n?n2:n);
      generic* b= new generic[m];
      for (i=0; i<k; i++) b[i]=a[i];
      if (l!=0) delete[] a;
      a=b;
    }
    else {
      if (l!=0) delete[] a;
      a= NULL;
    }
  }
  n=n2;
}

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

int
array_rep::Type () {
  return ARRAY_TYPE;
}

int
array_rep::Hash () {
  register int i, h=0;
  for (i=0; i<n; i++) {
    h = (h<<7) + (h>>25);
    h = h + hash (a[i]);
  }
  return h;
}

bool
array_rep::Equal (generic g) {
  if (gentype (g) != ARRAY_TYPE) return false;
  return array (this) == as_array (g);
}

generic
array_rep::Copy () {
  return copy (array (this));
}

int
array_rep::Arity () {
  return n;
}

generic&
array_rep::Access (int i) {
  return a[i];
}

generic
array_rep::Range (int begin, int end) {
  return generic (array (a + begin, end-begin));
}

generic&
array_rep::Car () {
  return a[0];
}

generic
array_rep::Cdr () {
  return Range (1, n);
}

void
array_rep::Print (ostream& out) {
  out << array (this);
}

/******************************************************************************
* Other routines on arrays
******************************************************************************/

array
copy (array a) {
  int i, n= N(a);
  array b (n);
  for (i=0; i<n; i++) b[n]= a[n];
  return b;
}

bool
operator == (array a, array b) {
  register int i;
  if (N(a)!=N(b)) return false;
  for (i=0; i<N(a); i++)
    if (a[i]!=b[i]) return false;
  return true;
}

bool
operator != (array a, array b) {
  register int i;
  if (N(a)!=N(b)) return true;
  for (i=0; i<N(a); i++)
    if (a[i]!=b[i]) return true;
  return false;
}

ostream&
operator << (ostream& out, array a) {
  out << "#array";
  print_entries (out, a);
  return out;
}

array&
operator << (array& a, generic g) {
  a->resize (N(a)+ 1);
  a[N(a)-1]=g;
  return a;
}

array&
operator << (array& a, array b) {
  register int i, k= N(a);
  a->resize (N(a)+ N(b));
  for (i=0; i<N(b); i++) a[i+k]= b[i];
  return a;
}

} // namespace mmx
