
/******************************************************************************
* MODULE     : mirror.hh
* DESCRIPTION: Mirror objects
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* A mirror of an object is a new object with the same "name".
* Mirrors can be quickly compared (like symbols),
* but several distinct mirrors of a same object can be created.
* This is useful for creating distinct type objects for
* types with the same name, but declared in a different scope.
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MIRROR_HH
#define __MIRROR_HH
#include <string.hh>
namespace mmx {

class mirror;
class mirror_rep: generic_rep {
protected:
  generic ref;

  int Type ();
  int Hash ();
  bool Equal (generic x);
  void Print (ostream& out);

public:
  inline mirror_rep (generic g): ref (g) {}

  friend class mirror;
  friend generic& S (mirror s);
  friend ostream& operator << (ostream& out, mirror s);
};

class mirror {
protected:
  mirror_rep* rep;
  inline mirror (mirror_rep* rep2): rep (rep2) {}

public:
  inline mirror (generic g):
    rep (new mirror_rep (g)) {}
  inline mirror (generic g, int begin, int end):
    rep (new mirror_rep (g)) { rep->index (begin, end); }
  inline mirror (generic g, generic where):
    rep (new mirror_rep (g)) { rep->index (where); }
  inline friend mirror as_mirror (generic g) {
    if (gentype(g) != MIRROR_TYPE) type_mismatch (g, "mirror");
    return mirror ((mirror_rep*) g.rep); }
  inline operator generic () { return generic (rep); }
  inline mirror_rep* operator -> () { return rep; }

  inline bool operator == (mirror s) { return rep == s.rep; }
  inline bool operator != (mirror s) { return rep != s.rep; }
  inline friend generic& S (mirror s) { return s->ref; }
  inline friend ostream& operator << (ostream& out, mirror s) {
#ifdef DEBUG_ON
    return out << s->ref << "~";
#else
    return out << s->ref;
#endif
  }

  friend class mirror_rep;
};

}
#endif // __MIRROR_HH
